-------------------------------------------------------------------------------
-- Copyright (C) 2000 Centre National de la Recherche Scientifique (CNRS)    --
--                                                                           --
-- This program is free software; you can redistribute it and/or modify      --
-- it under the terms of the GNU General Public License as published by      --
-- the Free Software Foundation; either version 2 of the License, or         --
-- (at your option) any later version.                                       --
-- As a special exception, if other files instantiate generics from this     --
-- unit, or you link this unit with other files to produce an executable,    --
-- this unit does not by itself cause the resulting executable to be         --
-- covered by the GNU General Public License. This exception does not        --
-- however invalidate any other reasons why the executable file might be     --
-- covered by the GNU Public License.                                        --
--                                                                           --
-- This program is distributed in the hope that it will be useful,           --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of            --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             --
-- GNU General Public License for more details.                              --
--                                                                           --
-- You should have received a copy of the GNU General Public License         --
-- along with this program; if not, write to the Free Software               --
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA --
--                                                                           --
--  Author: Duncan Sands (Duncan.Sands@math.u-psud.fr)                       --
--          Departement de Mathematiques, Batiment 425,                      --
--          Universite de Paris-XI, Orsay, France.                           --
--          http://topo.math.u-psud.fr/~sands                                --
-------------------------------------------------------------------------------

package body Ada_BLAS.Complex is

   ----------
   -- AMAX --
   ----------

   function ICAMAX (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Fortran_Integer;
   pragma Import (Fortran, ICAMAX, Name_Prepend & "icamax" & Name_Append);

   function IZAMAX (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Fortran_Integer;
   pragma Import (Fortran, IZAMAX, Name_Prepend & "izamax" & Name_Append);

   function AMAX (
     N    : Natural;
     X    : Vector_Type;
     INCX : Natural
   ) return Index_Type is
   begin
      if N = 0 then
         return X'First;
      elsif (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) +
              INCX * (
                Integer (
                  ICAMAX (Fortran_Integer (N), X, Fortran_Integer (INCX))
                ) - 1
              )
            );
         when Double =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) +
              INCX * (
                Integer (
                  IZAMAX (Fortran_Integer (N), X, Fortran_Integer (INCX))
                ) - 1
              )
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AMAX;

   function AMAX (X : Vector_Type) return Index_Type is
   begin
      if X'Length = 0 then
         return X'First;
      end if;

      case Precision is
         when Single =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) - 1 +
              Integer (
                ICAMAX (X'Length, X, 1)
              )
            );
         when Double =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) - 1 +
              Integer (
                IZAMAX (X'Length, X, 1)
              )
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AMAX;

   ----------
   -- ASUM --
   ----------

   function SCASUM (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, SCASUM, Name_Prepend & "scasum" & Name_Append);

   function DZASUM (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, DZASUM, Name_Prepend & "dzasum" & Name_Append);

   function ASUM (
     N    : Natural;
     X    : Vector_Type;
     INCX : Natural
   ) return Float_Type'Base is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SCASUM (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Double =>
            return DZASUM (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ASUM;

   function ASUM (X : Vector_Type) return Float_Type'Base is
   begin
      case Precision is
         when Single =>
            return SCASUM (X'Length, X, 1);
         when Double =>
            return DZASUM (X'Length, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ASUM;

   ----------
   -- AXPY --
   ----------

   procedure CAXPY (
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, CAXPY, Name_Prepend & "caxpy" & Name_Append);

   procedure ZAXPY (
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZAXPY, Name_Prepend & "zaxpy" & Name_Append);

   procedure AXPY (
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CAXPY (
              Fortran_Integer (N),
              ALPHA,
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            ZAXPY (
              Fortran_Integer (N),
              ALPHA,
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AXPY;

   procedure AXPY (
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     Y     : in out Vector_Type
   ) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CAXPY (X'Length, ALPHA, X, 1, Y, 1);
         when Double =>
            ZAXPY (X'Length, ALPHA, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AXPY;

   ----------
   -- COPY --
   ----------

   procedure CCOPY (
     N    : in     Fortran_Integer;
     X    : in     Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, CCOPY, Name_Prepend & "ccopy" & Name_Append);

   procedure ZCOPY (
     N    : in     Fortran_Integer;
     X    : in     Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, ZCOPY, Name_Prepend & "zcopy" & Name_Append);

   procedure COPY (
     N    : in     Natural;
     X    : in     Vector_Type;
     INCX : in     Integer;
     Y    : in out Vector_Type;
     INCY : in     Integer
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CCOPY (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            ZCOPY (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end COPY;

   procedure COPY (
     X : in     Vector_Type;
     Y : in out Vector_Type
   ) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CCOPY (X'Length, X, 1, Y, 1);
         when Double =>
            ZCOPY (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end COPY;

   ----------
   -- DOTC --
   ----------

   function CDOTC (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer;
     Y    : Vector_Type;
     INCY : Fortran_Integer
   ) return Complex_Type;
   pragma Import (Fortran, CDOTC, Name_Prepend & "cdotu" & Name_Append);

   function ZDOTC (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer;
     Y    : Vector_Type;
     INCY : Fortran_Integer
   ) return Complex_Type;
   pragma Import (Fortran, ZDOTC, Name_Prepend & "zdotu" & Name_Append);

   function DOTC (
     N    : Natural;
     X    : Vector_Type;
     INCX : Integer;
     Y    : Vector_Type;
     INCY : Integer
   ) return Complex_Type is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return CDOTC (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            return ZDOTC (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DOTC;

   function DOTC (X, Y : Vector_Type) return Complex_Type is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return CDOTC (X'Length, X, 1, Y, 1);
         when Double =>
            return ZDOTC (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DOTC;

   ----------
   -- DOTU --
   ----------

   function CDOTU (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer;
     Y    : Vector_Type;
     INCY : Fortran_Integer
   ) return Complex_Type;
   pragma Import (Fortran, CDOTU, Name_Prepend & "cdotu" & Name_Append);

   function ZDOTU (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer;
     Y    : Vector_Type;
     INCY : Fortran_Integer
   ) return Complex_Type;
   pragma Import (Fortran, ZDOTU, Name_Prepend & "zdotu" & Name_Append);

   function DOTU (
     N    : Natural;
     X    : Vector_Type;
     INCX : Integer;
     Y    : Vector_Type;
     INCY : Integer
   ) return Complex_Type is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return CDOTU (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            return ZDOTU (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DOTU;

   function DOTU (X, Y : Vector_Type) return Complex_Type is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return CDOTU (X'Length, X, 1, Y, 1);
         when Double =>
            return ZDOTU (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DOTU;

   ----------
   -- GBMV --
   ----------

   procedure CGBMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     KL    : in     Fortran_Integer;
     KU    : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, CGBMV, Name_Prepend & "cgbmv" & Name_Append);

   procedure ZGBMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     KL    : in     Fortran_Integer;
     KU    : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZGBMV, Name_Prepend & "zgbmv" & Name_Append);

   procedure GBMV (
     TRANS : in     Transpose_Type;
     M     : in     Natural;
     N     : in     Natural;
     KL    : in     Natural;
     KU    : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) <= KL + KU or A'Length (2) < N or
        INCX = 0 or INCY = 0 then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if (N - 1) * abs INCX >= X'Length or
              (M - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if (M - 1) * abs INCX >= X'Length or
              (N - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            ZGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GBMV;

   procedure GBMV (
     TRANS : in     Transpose_Type;
     KL    : in     Natural;
     KU    : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type
   ) is
      M : Fortran_Integer;
   begin
      if A'Length (1) /= KL + KU + 1 or KU >= A'Length (2) then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if X'Length /= A'Length (2) or KL >= Y'Length then
               raise Argument_Error;
            end if;
            M := Y'Length;
         when Transpose | Conjugate_Transpose =>
            if Y'Length /= A'Length (2) or KL >= X'Length then
               raise Argument_Error;
            end if;
            M := X'Length;
      end case;

      case Precision is
         when Single =>
            CGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => M,
              N     => A'Length (2),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            ZGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => M,
              N     => A'Length (2),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GBMV;

   ----------
   -- GEMM --
   ----------

   procedure CGEMM (
     TRANSA : in     Character_Set;
     TRANSB : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     K      : in     Fortran_Integer;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type;
     LDC    : in     Fortran_Integer
   );
   pragma Import (Fortran, CGEMM, Name_Prepend & "cgemm" & Name_Append);

   procedure ZGEMM (
     TRANSA : in     Character_Set;
     TRANSB : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     K      : in     Fortran_Integer;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type;
     LDC    : in     Fortran_Integer
   );
   pragma Import (Fortran, ZGEMM, Name_Prepend & "zgemm" & Name_Append);

   procedure GEMM (
     TRANSA : in     Transpose_Type;
     TRANSB : in     Transpose_Type;
     M      : in     Natural;
     N      : in     Natural;
     K      : in     Natural;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < M or C'Length (2) < N then
         raise Argument_Error;
      end if;

      case TRANSA is
         when None =>
            if A'Length (1) < M or A'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < M then
               raise Argument_Error;
            end if;
      end case;

      case TRANSB is
         when None =>
            if B'Length (1) < K or B'Length (2) < N then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if B'Length (1) < N or B'Length (2) < K then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              K      => Fortran_Integer (K),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Double =>
            ZGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              K      => Fortran_Integer (K),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMM;

   procedure GEMM (
     TRANSA : in     Transpose_Type;
     TRANSB : in     Transpose_Type;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type
   ) is
      K        : Fortran_Integer;
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 then
         raise Argument_Error;
      end if;

      case TRANSA is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;

      case TRANSB is
         when None =>
            if B'Length (1) /= K or B'Length (2) /= C'Length (2) then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if B'Length (2) /= K or B'Length (1) /= C'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => C'Length (1),
              N      => C'Length (2),
              K      => K,
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Double =>
            ZGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => C'Length (1),
              N      => C'Length (2),
              K      => K,
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMM;

   ----------
   -- GEMV --
   ----------

   procedure CGEMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, CGEMV, Name_Prepend & "cgemv" & Name_Append);

   procedure ZGEMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZGEMV, Name_Prepend & "zgemv" & Name_Append);

   procedure GEMV (
     TRANS : in     Transpose_Type;
     M     : in     Natural;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < M or A'Length (2) < N or
        INCX = 0 or INCY = 0 then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if (N - 1) * abs INCX >= X'Length or
              (M - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if (M - 1) * abs INCX >= X'Length or
              (N - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            ZGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMV;

   procedure GEMV (
     TRANS : in     Transpose_Type;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type
   ) is
   begin
      if A'Length (1) = 0 then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if X'Length /= A'Length (2) or Y'Length /= A'Length (1) then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if X'Length /= A'Length (1) or Y'Length /= A'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            ZGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMV;

   ----------
   -- GERC --
   ----------

   procedure CGERC (
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, CGERC, Name_Prepend & "cgerc" & Name_Append);

   procedure ZGERC (
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZGERC, Name_Prepend & "zgerc" & Name_Append);

   procedure GERC (
     M     : in     Natural;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < M or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (M - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CGERC (
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZGERC (
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GERC;

   procedure GERC (
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or X'Length /= A'Length (1) or
        Y'Length /= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CGERC (
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZGERC (
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GERC;

   ----------
   -- GERU --
   ----------

   procedure CGERU (
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, CGERU, Name_Prepend & "cgeru" & Name_Append);

   procedure ZGERU (
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZGERU, Name_Prepend & "zgeru" & Name_Append);

   procedure GERU (
     M     : in     Natural;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < M or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (M - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CGERU (
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZGERU (
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GERU;

   procedure GERU (
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or X'Length /= A'Length (1) or
        Y'Length /= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CGERU (
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZGERU (
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GERU;

   ----------
   -- HBMV --
   ----------

   procedure CHBMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, CHBMV, Name_Prepend & "chbmv" & Name_Append);

   procedure ZHBMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHBMV, Name_Prepend & "zhbmv" & Name_Append);

   procedure HBMV (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) <= K or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            ZHBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HBMV;

   procedure HBMV (
     UPLO  : in     Triangle_Type;
     K     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type
   ) is
   begin
      if A'Length (1) /= K + 1 or X'Length /= A'Length (2) or
        Y'Length /= A'Length (2) or K >= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            ZHBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HBMV;

   ----------
   -- HEMM --
   ----------

   procedure CHEMM (
     SIDE  : in     Character_Set;
     UPLO  : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, CHEMM, Name_Prepend & "chemm" & Name_Append);

   procedure ZHEMM (
     SIDE  : in     Character_Set;
     UPLO  : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHEMM, Name_Prepend & "zhemm" & Name_Append);

   procedure HEMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < M or C'Length (2) < N or
          B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CHEMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZHEMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HEMM;

   procedure HEMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        A'Length (1) /= A'Length (2) or
          B'Length (1) /= C'Length (1) or B'Length (2) /= C'Length (2) then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= C'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CHEMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => C'Length (1),
              N     => C'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZHEMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => C'Length (1),
              N     => C'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HEMM;

   ----------
   -- HEMV --
   ----------

   procedure CHEMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, CHEMV, Name_Prepend & "chemv" & Name_Append);

   procedure ZHEMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHEMV, Name_Prepend & "zhemv" & Name_Append);

   procedure HEMV (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHEMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            ZHEMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HEMV;

   procedure HEMV (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        A'Length (1) /= X'Length or A'Length (1) /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHEMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            ZHEMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HEMV;

   ---------
   -- HER --
   ---------

   procedure CHER (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, CHER, Name_Prepend & "cher" & Name_Append);

   procedure ZHER (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHER, Name_Prepend & "zher" & Name_Append);

   procedure HER (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHER (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZHER (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HER;

   procedure HER (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        X'Length /= A'Length (1) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHER (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZHER (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HER;

   ----------
   -- HER2 --
   ----------

   procedure CHER2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, CHER2, Name_Prepend & "cher2" & Name_Append);

   procedure ZHER2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHER2, Name_Prepend & "zher2" & Name_Append);

   procedure HER2 (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHER2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZHER2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HER2;

   procedure HER2 (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        X'Length /= A'Length (1) or Y'Length /= A'Length (1) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHER2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            ZHER2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HER2;

   ----------
   -- HERK --
   ----------

   procedure CHERK (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, CHERK, Name_Prepend & "cherk" & Name_Append);

   procedure ZHERK (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHERK, Name_Prepend & "zherk" & Name_Append);

   procedure HERK (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < N or C'Length (2) < N then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) < N or A'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CHERK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZHERK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HERK;

   procedure HERK (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
      K : Fortran_Integer;
   begin
      if A'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) /= C'Length (2) then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;

      case Precision is
         when Single =>
            CHERK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZHERK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HERK;

   -----------
   -- HER2K --
   -----------

   procedure CHER2K (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, CHER2K, Name_Prepend & "cher2k" & Name_Append);

   procedure ZHER2K (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHER2K, Name_Prepend & "zher2k" & Name_Append);

   procedure HER2K (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     B     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < N or C'Length (2) < N then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) < N or A'Length (2) < K or
              B'Length (1) < N or B'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < N or
              B'Length (1) < K or B'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CHER2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZHER2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HER2K;

   procedure HER2K (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     B     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
      K : Fortran_Integer;
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        A'Length (1) /= B'Length (1) or A'Length (2) /= B'Length (2) or
          C'Length (1) /= C'Length (2) then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;

      case Precision is
         when Single =>
            CHER2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZHER2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HER2K;

   ----------
   -- HPMV --
   ----------

   procedure CHPMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, CHPMV, Name_Prepend & "chpmv" & Name_Append);

   procedure ZHPMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZHPMV, Name_Prepend & "zhpmv" & Name_Append);

   procedure HPMV (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or INCY = 0 or
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            ZHPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HPMV;

   procedure HPMV (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Complex_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     BETA  : in     Complex_Type;
     Y     : in out Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (1 + X'Length) / 2 or
        X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            ZHPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HPMV;

   ---------
   -- HPR --
   ---------

   procedure CHPR (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, CHPR, Name_Prepend & "chpr" & Name_Append);

   procedure ZHPR (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, ZHPR, Name_Prepend & "zhpr" & Name_Append);

   procedure HPR (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or
        (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              AP    => AP
            );
         when Double =>
            ZHPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HPR;

   procedure HPR (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (X'Length + 1) / 2 then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              AP    => AP
            );
         when Double =>
            ZHPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HPR;

   ----------
   -- HPR2 --
   ----------

   procedure CHPR2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, CHPR2, Name_Prepend & "chpr2" & Name_Append);

   procedure ZHPR2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, ZHPR2, Name_Prepend & "zhpr2" & Name_Append);

   procedure HPR2 (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or INCY = 0 or
        (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              AP    => AP
            );
         when Double =>
            ZHPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HPR2;

   procedure HPR2 (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Complex_Type;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (X'Length + 1) / 2 or
        X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CHPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              AP    => AP
            );
         when Double =>
            ZHPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end HPR2;

   ----------
   -- NRM2 --
   ----------

   function SCNRM2 (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, SCNRM2, Name_Prepend & "scnrm2" & Name_Append);

   function DZNRM2 (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, DZNRM2, Name_Prepend & "dznrm2" & Name_Append);

   function NRM2 (
     N    : Natural;
     X    : Vector_Type;
     INCX : Natural
   ) return Float_Type'Base is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SCNRM2 (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Double =>
            return DZNRM2 (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end NRM2;

   function NRM2 (X : Vector_Type) return Float_Type'Base is
   begin
      case Precision is
         when Single =>
            return SCNRM2 (X'Length, X, 1);
         when Double =>
            return DZNRM2 (X'Length, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end NRM2;

   ----------
   -- SCAL --
   ----------

   procedure CSCAL (
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CSCAL, Name_Prepend & "cscal" & Name_Append);

   procedure CSSCAL (
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CSSCAL, Name_Prepend & "csscal" & Name_Append);

   procedure ZSCAL (
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZSCAL, Name_Prepend & "zscal" & Name_Append);

   procedure ZDSCAL (
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZDSCAL, Name_Prepend & "zdscal" & Name_Append);

   procedure SCAL (
     N     : in     Natural;
     ALPHA : in     Complex_Type;
     X     : in out Vector_Type;
     INCX  : in     Natural
   ) is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CSCAL (Fortran_Integer (N), ALPHA, X, Fortran_Integer (INCX));
         when Double =>
            ZSCAL (Fortran_Integer (N), ALPHA, X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SCAL;

   procedure SCAL (
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type;
     INCX  : in     Natural
   ) is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CSSCAL (Fortran_Integer (N), ALPHA, X, Fortran_Integer (INCX));
         when Double =>
            ZDSCAL (Fortran_Integer (N), ALPHA, X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SCAL;

   procedure SCAL (
     ALPHA : in     Complex_Type;
     X     : in out Vector_Type
   ) is
   begin
      case Precision is
         when Single =>
            CSCAL (X'Length, ALPHA, X, 1);
         when Double =>
            ZSCAL (X'Length, ALPHA, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SCAL;

   procedure SCAL (
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type
   ) is
   begin
      case Precision is
         when Single =>
            CSSCAL (X'Length, ALPHA, X, 1);
         when Double =>
            ZDSCAL (X'Length, ALPHA, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SCAL;

   ----------
   -- SWAP --
   ----------

   procedure CSWAP (
     N    : in     Fortran_Integer;
     X    : in out Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, CSWAP, Name_Prepend & "cswap" & Name_Append);

   procedure ZSWAP (
     N    : in     Fortran_Integer;
     X    : in out Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, ZSWAP, Name_Prepend & "zswap" & Name_Append);

   procedure SWAP (
     N    : in     Natural;
     X    : in out Vector_Type;
     INCX : in     Integer;
     Y    : in out Vector_Type;
     INCY : in     Integer
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CSWAP (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            ZSWAP (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SWAP;

   procedure SWAP (X, Y : in out Vector_Type) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CSWAP (X'Length, X, 1, Y, 1);
         when Double =>
            ZSWAP (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SWAP;

   ----------
   -- SYMM --
   ----------

   procedure CSYMM (
     SIDE  : in     Character_Set;
     UPLO  : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, CSYMM, Name_Prepend & "csymm" & Name_Append);

   procedure ZSYMM (
     SIDE  : in     Character_Set;
     UPLO  : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZSYMM, Name_Prepend & "zsymm" & Name_Append);

   procedure SYMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < M or C'Length (2) < N or
          B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYMM;

   procedure SYMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Complex_Type;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        A'Length (1) /= A'Length (2) or
          B'Length (1) /= C'Length (1) or B'Length (2) /= C'Length (2) then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= C'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => C'Length (1),
              N     => C'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => C'Length (1),
              N     => C'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYMM;

   ----------
   -- SYRK --
   ----------

   procedure CSYRK (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, CSYRK, Name_Prepend & "csyrk" & Name_Append);

   procedure ZSYRK (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZSYRK, Name_Prepend & "zsyrk" & Name_Append);

   procedure SYRK (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < N or C'Length (2) < N then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) < N or A'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYRK;

   procedure SYRK (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
      K : Fortran_Integer;
   begin
      if A'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) /= C'Length (2) then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;

      case Precision is
         when Single =>
            CSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYRK;

   -----------
   -- SYR2K --
   -----------

   procedure CSYR2K (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, CSYR2K, Name_Prepend & "csyr2k" & Name_Append);

   procedure ZSYR2K (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, ZSYR2K, Name_Prepend & "zsyr2k" & Name_Append);

   procedure SYR2K (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     B     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < N or C'Length (2) < N then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) < N or A'Length (2) < K or
              B'Length (1) < N or B'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < N or
              B'Length (1) < K or B'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR2K;

   procedure SYR2K (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     ALPHA : in     Complex_Type;
     A     : in     Matrix_Type;
     B     : in     Matrix_Type;
     BETA  : in     Complex_Type;
     C     : in out Matrix_Type
   ) is
      K : Fortran_Integer;
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        A'Length (1) /= B'Length (1) or A'Length (2) /= B'Length (2) or
          C'Length (1) /= C'Length (2) then
         raise Argument_Error;
      end if;

      case TRANS is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;

      case Precision is
         when Single =>
            CSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            ZSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR2K;

   ----------
   -- TBMV --
   ----------

   procedure CTBMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CTBMV, Name_Prepend & "ctbmv" & Name_Append);

   procedure ZTBMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTBMV, Name_Prepend & "ztbmv" & Name_Append);

   procedure TBMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) <= K or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            ZTBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBMV;

   procedure TBMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) /= K + 1 or X'Length /= A'Length (2) or
        K >= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            ZTBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBMV;

   ----------
   -- TBSV --
   ----------

   procedure CTBSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CTBSV, Name_Prepend & "ctbsv" & Name_Append);

   procedure ZTBSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTBSV, Name_Prepend & "ztbsv" & Name_Append);

   procedure TBSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) <= K or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            ZTBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBSV;

   procedure TBSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) /= K + 1 or X'Length /= A'Length (2) or
        K >= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            ZTBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBSV;

   ----------
   -- TPMV --
   ----------

   procedure CTPMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CTPMV, Name_Prepend & "ctpmv" & Name_Append);

   procedure ZTPMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTPMV, Name_Prepend & "ztpmv" & Name_Append);

   procedure TPMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or
        (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            ZTPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPMV;

   procedure TPMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (1 + X'Length) / 2 then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Double =>
            ZTPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPMV;

   ----------
   -- TPSV --
   ----------

   procedure CTPSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CTPSV, Name_Prepend & "ctpsv" & Name_Append);

   procedure ZTPSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTPSV, Name_Prepend & "ztpsv" & Name_Append);

   procedure TPSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or
        (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            ZTPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPSV;

   procedure TPSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (1 + X'Length) / 2 then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Double =>
            ZTPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPSV;

   ----------
   -- TRMM --
   ----------

   procedure CTRMM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, CTRMM, Name_Prepend & "ctrmm" & Name_Append);

   procedure ZTRMM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTRMM, Name_Prepend & "ztrmm" & Name_Append);

   procedure TRMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CTRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            ZTRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMM;

   procedure TRMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        A'Length (1) /= A'Length (2) then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) /= B'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= B'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CTRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            ZTRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMM;

   ----------
   -- TRMV --
   ----------

   procedure CTRMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CTRMV, Name_Prepend & "ctrmv" & Name_Append);

   procedure ZTRMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTRMV, Name_Prepend & "ztrmv" & Name_Append);

   procedure TRMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            ZTRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMV;

   procedure TRMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        A'Length (1) /= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            ZTRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMV;

   ----------
   -- TRSM --
   ----------

   procedure CTRSM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, CTRSM, Name_Prepend & "ctrsm" & Name_Append);

   procedure ZTRSM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTRSM, Name_Prepend & "ztrsm" & Name_Append);

   procedure TRSM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CTRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            ZTRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSM;

   procedure TRSM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     ALPHA  : in     Complex_Type;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        A'Length (1) /= A'Length (2) then
         raise Argument_Error;
      end if;

      case SIDE is
         when Left =>
            if A'Length (1) /= B'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= B'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            CTRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            ZTRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSM;

   ----------
   -- TRSV --
   ----------

   procedure CTRSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, CTRSV, Name_Prepend & "ctrsv" & Name_Append);

   procedure ZTRSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, ZTRSV, Name_Prepend & "ztrsv" & Name_Append);

   procedure TRSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            ZTRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSV;

   procedure TRSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        A'Length (1) /= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            CTRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            ZTRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSV;

end Ada_BLAS.Complex;

