-------------------------------------------------------------------------------
-- Copyright (C) 2000 Centre National de la Recherche Scientifique (CNRS)    --
--                                                                           --
-- This program is free software; you can redistribute it and/or modify      --
-- it under the terms of the GNU General Public License as published by      --
-- the Free Software Foundation; either version 2 of the License, or         --
-- (at your option) any later version.                                       --
-- As a special exception, if other files instantiate generics from this     --
-- unit, or you link this unit with other files to produce an executable,    --
-- this unit does not by itself cause the resulting executable to be         --
-- covered by the GNU General Public License. This exception does not        --
-- however invalidate any other reasons why the executable file might be     --
-- covered by the GNU Public License.                                        --
--                                                                           --
-- This program is distributed in the hope that it will be useful,           --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of            --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             --
-- GNU General Public License for more details.                              --
--                                                                           --
-- You should have received a copy of the GNU General Public License         --
-- along with this program; if not, write to the Free Software               --
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA --
--                                                                           --
--  Author: Duncan Sands (Duncan.Sands@math.u-psud.fr)                       --
--          Departement de Mathematiques, Batiment 425,                      --
--          Universite de Paris-XI, Orsay, France.                           --
--          http://topo.math.u-psud.fr/~sands                                --
-------------------------------------------------------------------------------

package body Ada_BLAS.Real is

   ----------
   -- AMAX --
   ----------

   function ISAMAX (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Fortran_Integer;
   pragma Import (Fortran, ISAMAX, Name_Prepend & "isamax" & Name_Append);

   function IDAMAX (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Fortran_Integer;
   pragma Import (Fortran, IDAMAX, Name_Prepend & "idamax" & Name_Append);

   function AMAX (
     N    : Natural;
     X    : Vector_Type;
     INCX : Natural
   ) return Index_Type is
   begin
      if N = 0 then
         return X'First;
      elsif (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) +
              INCX * (
                Integer (
                  ISAMAX (Fortran_Integer (N), X, Fortran_Integer (INCX))
                ) - 1
              )
            );
         when Double =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) +
              INCX * (
                Integer (
                  IDAMAX (Fortran_Integer (N), X, Fortran_Integer (INCX))
                ) - 1
              )
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AMAX;

   function AMAX (X : Vector_Type) return Index_Type is
   begin
      if X'Length = 0 then
         return X'First;
      end if;
      
      case Precision is
         when Single =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) - 1 +
              Integer (
                ISAMAX (X'Length, X, 1)
              )
            );
         when Double =>
            return Index_Type'Val (
              Index_Type'Pos (X'First) - 1 +
              Integer (
                IDAMAX (X'Length, X, 1)
              )
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AMAX;

   ----------
   -- ASUM --
   ----------

   function SASUM (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, SASUM, Name_Prepend & "sasum" & Name_Append);

   function DASUM (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, DASUM, Name_Prepend & "dasum" & Name_Append);

   function ASUM (
     N    : Natural;
     X    : Vector_Type;
     INCX : Natural
   ) return Float_Type'Base is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SASUM (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Double =>
            return DASUM (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ASUM;

   function ASUM (X : Vector_Type) return Float_Type'Base is
   begin
      case Precision is
         when Single =>
            return SASUM (X'Length, X, 1);
         when Double =>
            return DASUM (X'Length, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ASUM;

   ----------
   -- AXPY --
   ----------

   procedure SAXPY (
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, SAXPY, Name_Prepend & "saxpy" & Name_Append);

   procedure DAXPY (
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, DAXPY, Name_Prepend & "daxpy" & Name_Append);

   procedure AXPY (
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SAXPY (
              Fortran_Integer (N),
              ALPHA,
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            DAXPY (
              Fortran_Integer (N),
              ALPHA,
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AXPY;

   procedure AXPY (
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     Y     : in out Vector_Type
   ) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SAXPY (X'Length, ALPHA, X, 1, Y, 1);
         when Double =>
            DAXPY (X'Length, ALPHA, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end AXPY;

   ----------
   -- COPY --
   ----------

   procedure SCOPY (
     N    : in     Fortran_Integer;
     X    : in     Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, SCOPY, Name_Prepend & "scopy" & Name_Append);

   procedure DCOPY (
     N    : in     Fortran_Integer;
     X    : in     Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, DCOPY, Name_Prepend & "dcopy" & Name_Append);

   procedure COPY (
     N    : in     Natural;
     X    : in     Vector_Type;
     INCX : in     Integer;
     Y    : in out Vector_Type;
     INCY : in     Integer
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SCOPY (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            DCOPY (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end COPY;

   procedure COPY (
     X : in     Vector_Type;
     Y : in out Vector_Type
   ) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SCOPY (X'Length, X, 1, Y, 1);
         when Double =>
            DCOPY (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end COPY;

   ---------
   -- DOT --
   ---------

   function SDOT (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer;
     Y    : Vector_Type;
     INCY : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, SDOT, Name_Prepend & "sdot" & Name_Append);

   function DDOT (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer;
     Y    : Vector_Type;
     INCY : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, DDOT, Name_Prepend & "ddot" & Name_Append);

   function DOT (
     N    : Natural;
     X    : Vector_Type;
     INCX : Integer;
     Y    : Vector_Type;
     INCY : Integer
   ) return Float_Type'Base is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SDOT (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            return DDOT (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DOT;

   function DOT (X, Y : Vector_Type) return Float_Type'Base is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SDOT (X'Length, X, 1, Y, 1);
         when Double =>
            return DDOT (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DOT;

   -----------
   -- DSDOT --
   -----------

   function SDSDOT (
     N     : Fortran_Integer;
     ALPHA : Float_Type'Base;
     X     : Vector_Type;
     INCX  : Fortran_Integer;
     Y     : Vector_Type;
     INCY  : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, SDSDOT, Name_Prepend & "sdsdot" & Name_Append);

   function DSDOT (
     N     : Natural;
     ALPHA : Float_Type'Base;
     X     : Vector_Type;
     INCX  : Integer;
     Y     : Vector_Type;
     INCY  : Integer
   ) return Float_Type'Base is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SDSDOT (
              Fortran_Integer (N),
              ALPHA,
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            return ALPHA + DDOT (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DSDOT;

   function DSDOT (
     ALPHA : Float_Type'Base;
     X, Y  : Vector_Type
   ) return Float_Type'Base is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SDSDOT (X'Length, ALPHA, X, 1, Y, 1);
         when Double =>
            return ALPHA + DDOT (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end DSDOT;

   ----------
   -- GBMV --
   ----------

   procedure SGBMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     KL    : in     Fortran_Integer;
     KU    : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, SGBMV, Name_Prepend & "sgbmv" & Name_Append);

   procedure DGBMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     KL    : in     Fortran_Integer;
     KU    : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, DGBMV, Name_Prepend & "dgbmv" & Name_Append);

   procedure GBMV (
     TRANS : in     Transpose_Type;
     M     : in     Natural;
     N     : in     Natural;
     KL    : in     Natural;
     KU    : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) <= KL + KU or A'Length (2) < N or
        INCX = 0 or INCY = 0 then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if (N - 1) * abs INCX >= X'Length or
              (M - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if (M - 1) * abs INCX >= X'Length or
              (N - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            SGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            DGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GBMV;

   procedure GBMV (
     TRANS : in     Transpose_Type;
     KL    : in     Natural;
     KU    : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type
   ) is
      M : Fortran_Integer;
   begin
      if A'Length (1) /= KL + KU + 1 or KU >= A'Length (2) then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if X'Length /= A'Length (2) or KL >= Y'Length then
               raise Argument_Error;
            end if;
            M := Y'Length;
         when Transpose | Conjugate_Transpose =>
            if Y'Length /= A'Length (2) or KL >= X'Length then
               raise Argument_Error;
            end if;
            M := X'Length;
      end case;

      case Precision is
         when Single =>
            SGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => M,
              N     => A'Length (2),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            DGBMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => M,
              N     => A'Length (2),
              KL    => Fortran_Integer (KL),
              KU    => Fortran_Integer (KU),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GBMV;

   ----------
   -- GEMM --
   ----------

   procedure SGEMM (
     TRANSA : in     Character_Set;
     TRANSB : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     K      : in     Fortran_Integer;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer;
     BETA   : in     Float_Type'Base;
     C      : in out Matrix_Type;
     LDC    : in     Fortran_Integer
   );
   pragma Import (Fortran, SGEMM, Name_Prepend & "sgemm" & Name_Append);

   procedure DGEMM (
     TRANSA : in     Character_Set;
     TRANSB : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     K      : in     Fortran_Integer;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer;
     BETA   : in     Float_Type'Base;
     C      : in out Matrix_Type;
     LDC    : in     Fortran_Integer
   );
   pragma Import (Fortran, DGEMM, Name_Prepend & "dgemm" & Name_Append);
   
   procedure GEMM (
     TRANSA : in     Transpose_Type;
     TRANSB : in     Transpose_Type;
     M      : in     Natural;
     N      : in     Natural;
     K      : in     Natural;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Float_Type'Base;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < M or C'Length (2) < N then
         raise Argument_Error;
      end if;
      
      case TRANSA is
         when None =>
            if A'Length (1) < M or A'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < M then
               raise Argument_Error;
            end if;
      end case;

      case TRANSB is
         when None =>
            if B'Length (1) < K or B'Length (2) < N then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if B'Length (1) < N or B'Length (2) < K then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            SGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              K      => Fortran_Integer (K),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Double =>
            DGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              K      => Fortran_Integer (K),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMM;

   procedure GEMM (
     TRANSA : in     Transpose_Type;
     TRANSB : in     Transpose_Type;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Float_Type'Base;
     C      : in out Matrix_Type
   ) is
      K        : Fortran_Integer;
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 then
         raise Argument_Error;
      end if;
      
      case TRANSA is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;

      case TRANSB is
         when None =>
            if B'Length (1) /= K or B'Length (2) /= C'Length (2) then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if B'Length (2) /= K or B'Length (1) /= C'Length (2) then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            SGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => C'Length (1),
              N      => C'Length (2),
              K      => K,
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Double =>
            DGEMM (
              TRANSA => Fortran_TRANS (TRANSA),
              TRANSB => Fortran_TRANS (TRANSB),
              M      => C'Length (1),
              N      => C'Length (2),
              K      => K,
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1),
              BETA   => BETA,
              C      => C,
              LDC    => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMM;

   ----------
   -- GEMV --
   ----------

   procedure SGEMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, SGEMV, Name_Prepend & "sgemv" & Name_Append);

   procedure DGEMV (
     TRANS : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, DGEMV, Name_Prepend & "dgemv" & Name_Append);

   procedure GEMV (
     TRANS : in     Transpose_Type;
     M     : in     Natural;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < M or A'Length (2) < N or
        INCX = 0 or INCY = 0 then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if (N - 1) * abs INCX >= X'Length or
              (M - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if (M - 1) * abs INCX >= X'Length or
              (N - 1) * abs INCY >= Y'Length then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            SGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            DGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMV;

   procedure GEMV (
     TRANS : in     Transpose_Type;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type
   ) is
   begin
      if A'Length (1) = 0 then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if X'Length /= A'Length (2) or Y'Length /= A'Length (1) then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if X'Length /= A'Length (1) or Y'Length /= A'Length (2) then
               raise Argument_Error;
            end if;
      end case;

      case Precision is
         when Single =>
            SGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            DGEMV (
              TRANS => Fortran_TRANS (TRANS),
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GEMV;

   ---------
   -- GER --
   ---------

   procedure SGER (
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, SGER, Name_Prepend & "sger" & Name_Append);

   procedure DGER (
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, DGER, Name_Prepend & "dger" & Name_Append);

   procedure GER (
     M     : in     Natural;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < M or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (M - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SGER (
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            DGER (
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GER;

   procedure GER (
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or X'Length /= A'Length (1) or
        Y'Length /= A'Length (2) then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SGER (
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            DGER (
              M     => A'Length (1),
              N     => A'Length (2),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end GER;

   ----------
   -- NRM2 --
   ----------

   function SNRM2 (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, SNRM2, Name_Prepend & "snrm2" & Name_Append);

   function DNRM2 (
     N    : Fortran_Integer;
     X    : Vector_Type;
     INCX : Fortran_Integer
   ) return Float_Type'Base;
   pragma Import (Fortran, DNRM2, Name_Prepend & "dnrm2" & Name_Append);

   function NRM2 (
     N    : Natural;
     X    : Vector_Type;
     INCX : Natural
   ) return Float_Type'Base is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            return SNRM2 (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Double =>
            return DNRM2 (Fortran_Integer (N), X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end NRM2;

   function NRM2 (X : Vector_Type) return Float_Type'Base is
   begin
      case Precision is
         when Single =>
            return SNRM2 (X'Length, X, 1);
         when Double =>
            return DNRM2 (X'Length, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end NRM2;

   ---------
   -- ROT --
   ---------

   procedure SROT (
     N    : in     Fortran_Integer;
     X    : in out Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer;
     C, S : in     Float_Type'Base
   );
   pragma Import (Fortran, SROT, Name_Prepend & "srot" & Name_Append);

   procedure DROT (
     N    : in     Fortran_Integer;
     X    : in out Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer;
     C, S : in     Float_Type'Base
   );
   pragma Import (Fortran, DROT, Name_Prepend & "drot" & Name_Append);

   procedure ROT (
     N    : in     Natural;
     X    : in out Vector_Type;
     INCX : in     Integer;
     Y    : in out Vector_Type;
     INCY : in     Integer;
     C, S : in     Float_Type'Base
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SROT (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY),
              C, S
            );
         when Double =>
            DROT (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY),
              C, S
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ROT;

   procedure ROT (
     X, Y : in out Vector_Type;
     C, S : in     Float_Type'Base
   ) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SROT (X'Length, X, 1, Y, 1, C, S);
         when Double =>
            DROT (X'Length, X, 1, Y, 1, C, S);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ROT;

   ----------
   -- ROTM --
   ----------

   procedure SROTM (
     N     : in     Fortran_Integer;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer;
     PARAM : in     Modified_Givens
   );
   pragma Import (Fortran, SROTM, Name_Prepend & "srotm" & Name_Append);

   procedure DROTM (
     N     : in     Fortran_Integer;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer;
     PARAM : in     Modified_Givens
   );
   pragma Import (Fortran, DROTM, Name_Prepend & "drotm" & Name_Append);

   procedure ROTM (
     N     : in     Natural;
     X     : in out Vector_Type;
     INCX  : in     Integer;
     Y     : in out Vector_Type;
     INCY  : in     Integer;
     PARAM : in     Modified_Givens
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SROTM (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY),
              PARAM
            );
         when Double =>
            DROTM (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY),
              PARAM
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ROTM;

   procedure ROTM (
     X, Y  : in out Vector_Type;
     PARAM : in     Modified_Givens
   ) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SROTM (X'Length, X, 1, Y, 1, PARAM);
         when Double =>
            DROTM (X'Length, X, 1, Y, 1, PARAM);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ROTM;

   ----------
   -- ROTG --
   ----------

   procedure SROTG (
     A, B : in out Float_Type'Base;
     C, S :    out Float_Type'Base
   );
   pragma Import (Fortran, SROTG, Name_Prepend & "srotg" & Name_Append);

   procedure DROTG (
     A, B : in out Float_Type'Base;
     C, S :    out Float_Type'Base
   );
   pragma Import (Fortran, DROTG, Name_Prepend & "drotg" & Name_Append);

   procedure ROTG (
     A, B : in out Float_Type'Base;
     C, S :    out Float_Type'Base
   ) is
   begin
      case Precision is
         when Single =>
            SROTG (A, B, C, S);
         when Double =>
            DROTG (A, B, C, S);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ROTG;

   -----------
   -- ROTMG --
   -----------

   procedure SROTMG (
     D1, D2 : in out Float_Type'Base;
     A      : in out Float_Type'Base;
     B      : in     Float_Type'Base;
     PARAM  :    out Modified_Givens
   );
   pragma Import (Fortran, SROTMG, Name_Prepend & "srotmg" & Name_Append);

   procedure DROTMG (
     D1, D2 : in out Float_Type'Base;
     A      : in out Float_Type'Base;
     B      : in     Float_Type'Base;
     PARAM  :    out Modified_Givens
   );
   pragma Import (Fortran, DROTMG, Name_Prepend & "drotmg" & Name_Append);

   procedure ROTMG (
     D1, D2 : in out Float_Type'Base;
     A      : in out Float_Type'Base;
     B      : in     Float_Type'Base;
     PARAM  :    out Modified_Givens
   ) is
   begin
      case Precision is
         when Single =>
            SROTMG (D1, D2, A, B, PARAM);
         when Double =>
            DROTMG (D1, D2, A, B, PARAM);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end ROTMG;

   ----------
   -- SBMV --
   ----------

   procedure SSBMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, SSBMV, Name_Prepend & "ssbmv" & Name_Append);

   procedure DSBMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, DSBMV, Name_Prepend & "dsbmv" & Name_Append);

   procedure SBMV (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) <= K or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            DSBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SBMV;

   procedure SBMV (
     UPLO  : in     Triangle_Type;
     K     : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type
   ) is
   begin
      if A'Length (1) /= K + 1 or X'Length /= A'Length (2) or
        Y'Length /= A'Length (2) or K >= A'Length (2) then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            DSBMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SBMV;

   ----------
   -- SCAL --
   ----------

   procedure SSCAL (
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, SSCAL, Name_Prepend & "sscal" & Name_Append);

   procedure DSCAL (
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DSCAL, Name_Prepend & "dscal" & Name_Append);

   procedure SCAL (
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type;
     INCX  : in     Natural
   ) is
   begin
      if N > 0 and (N - 1) * INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSCAL (Fortran_Integer (N), ALPHA, X, Fortran_Integer (INCX));
         when Double =>
            DSCAL (Fortran_Integer (N), ALPHA, X, Fortran_Integer (INCX));
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SCAL;

   procedure SCAL (
     ALPHA : in     Float_Type'Base;
     X     : in out Vector_Type
   ) is
   begin
      case Precision is
         when Single =>
            SSCAL (X'Length, ALPHA, X, 1);
         when Double =>
            DSCAL (X'Length, ALPHA, X, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SCAL;

   ----------
   -- SPMV --
   ----------

   procedure SSPMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, SSPMV, Name_Prepend & "sspmv" & Name_Append);

   procedure DSPMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, DSPMV, Name_Prepend & "dspmv" & Name_Append);

   procedure SPMV (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or INCY = 0 or
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            DSPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SPMV;

   procedure SPMV (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Float_Type'Base;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (1 + X'Length) / 2 or
        X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            DSPMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              AP    => AP,
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SPMV;

   ---------
   -- SPR --
   ---------

   procedure SSPR (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, SSPR, Name_Prepend & "sspr" & Name_Append);

   procedure DSPR (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, DSPR, Name_Prepend & "dspr" & Name_Append);

   procedure SPR (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or
        (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              AP    => AP
            );
         when Double =>
            DSPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SPR;

   procedure SPR (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (X'Length + 1) / 2 then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              AP    => AP
            );
         when Double =>
            DSPR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SPR;

   ----------
   -- SPR2 --
   ----------

   procedure SSPR2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, SSPR2, Name_Prepend & "sspr2" & Name_Append);

   procedure DSPR2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     AP    : in out Vector_Type
   );
   pragma Import (Fortran, DSPR2, Name_Prepend & "dspr2" & Name_Append);

   procedure SPR2 (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or INCY = 0 or
        (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              AP    => AP
            );
         when Double =>
            DSPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SPR2;

   procedure SPR2 (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     AP    : in out Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (X'Length + 1) / 2 or
        X'Length /= Y'Length then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              AP    => AP
            );
         when Double =>
            DSPR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => X'Length,
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              AP    => AP
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SPR2;

   ----------
   -- SWAP --
   ----------

   procedure SSWAP (
     N    : in     Fortran_Integer;
     X    : in out Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, SSWAP, Name_Prepend & "sswap" & Name_Append);

   procedure DSWAP (
     N    : in     Fortran_Integer;
     X    : in out Vector_Type;
     INCX : in     Fortran_Integer;
     Y    : in out Vector_Type;
     INCY : in     Fortran_Integer
   );
   pragma Import (Fortran, DSWAP, Name_Prepend & "dswap" & Name_Append);

   procedure SWAP (
     N    : in     Natural;
     X    : in out Vector_Type;
     INCX : in     Integer;
     Y    : in out Vector_Type;
     INCY : in     Integer
   ) is
   begin
      if N > 0 and (
        (N - 1) * abs INCX >= X'Length or (N - 1) * abs INCY >= Y'Length
      ) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSWAP (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Double =>
            DSWAP (
              Fortran_Integer (N),
              X,
              Fortran_Integer (INCX),
              Y,
              Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SWAP;

   procedure SWAP (X, Y : in out Vector_Type) is
   begin
      if X'Length /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSWAP (X'Length, X, 1, Y, 1);
         when Double =>
            DSWAP (X'Length, X, 1, Y, 1);
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SWAP;
   
   ----------
   -- SYMM --
   ----------

   procedure SSYMM (
     SIDE  : in     Character_Set;
     UPLO  : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, SSYMM, Name_Prepend & "ssymm" & Name_Append);

   procedure DSYMM (
     SIDE  : in     Character_Set;
     UPLO  : in     Character_Set;
     M     : in     Fortran_Integer;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, DSYMM, Name_Prepend & "dsymm" & Name_Append);

   procedure SYMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Float_Type'Base;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < M or C'Length (2) < N or
          B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;
      
      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            SSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            DSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => Fortran_Integer (M),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYMM;

   procedure SYMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type;
     BETA   : in     Float_Type'Base;
     C      : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        A'Length (1) /= A'Length (2) or
          B'Length (1) /= C'Length (1) or B'Length (2) /= C'Length (2) then
         raise Argument_Error;
      end if;
      
      case SIDE is
         when Left =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= C'Length (2) then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            SSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => C'Length (1),
              N     => C'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            DSYMM (
              SIDE  => Fortran_SIDE (SIDE),
              UPLO  => Fortran_UPLO (UPLO),
              M     => C'Length (1),
              N     => C'Length (2),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYMM;

   ----------
   -- SYMV --
   ----------

   procedure SSYMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, SSYMV, Name_Prepend & "ssymv" & Name_Append);

   procedure DSYMV (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Fortran_Integer
   );
   pragma Import (Fortran, DSYMV, Name_Prepend & "dsymv" & Name_Append);

   procedure SYMV (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type;
     INCY  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSYMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Double =>
            DSYMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX),
              BETA  => BETA,
              Y     => Y,
              INCY  => Fortran_Integer (INCY)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYMV;

   procedure SYMV (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     BETA  : in     Float_Type'Base;
     Y     : in out Vector_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        A'Length (1) /= X'Length or A'Length (1) /= Y'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            SSYMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Double =>
            DSYMV (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1,
              BETA  => BETA,
              Y     => Y,
              INCY  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYMV;

   ---------
   -- SYR --
   ---------

   procedure SSYR (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, SSYR, Name_Prepend & "ssyr" & Name_Append);

   procedure DSYR (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, DSYR, Name_Prepend & "dsyr" & Name_Append);

   procedure SYR (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSYR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            DSYR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR;

   procedure SYR (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        X'Length /= A'Length (1) then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSYR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            DSYR (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR;

   ----------
   -- SYR2 --
   ----------

   procedure SSYR2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, SSYR2, Name_Prepend & "ssyr2" & Name_Append);

   procedure DSYR2 (
     UPLO  : in     Character_Set;
     N     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer;
     Y     : in     Vector_Type;
     INCY  : in     Fortran_Integer;
     A     : in out Matrix_Type;
     LDA   : in     Fortran_Integer
   );
   pragma Import (Fortran, DSYR2, Name_Prepend & "dsyr2" & Name_Append);

   procedure SYR2 (
     UPLO  : in     Triangle_Type;
     N     : in     Natural;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     INCX  : in     Integer;
     Y     : in     Vector_Type;
     INCY  : in     Integer;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or INCY = 0 or (N - 1) * abs INCX >= X'Length or
          (N - 1) * abs INCY >= Y'Length then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSYR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            DSYR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => Fortran_Integer (N),
              ALPHA => ALPHA,
              X     => X,
              INCX  => Fortran_Integer (INCX),
              Y     => Y,
              INCY  => Fortran_Integer (INCY),
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR2;

   procedure SYR2 (
     UPLO  : in     Triangle_Type;
     ALPHA : in     Float_Type'Base;
     X     : in     Vector_Type;
     Y     : in     Vector_Type;
     A     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        X'Length /= A'Length (1) or Y'Length /= A'Length (1) then
         raise Argument_Error;
      end if;
      
      case Precision is
         when Single =>
            SSYR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Double =>
            DSYR2 (
              UPLO  => Fortran_UPLO (UPLO),
              N     => A'Length (1),
              ALPHA => ALPHA,
              X     => X,
              INCX  => 1,
              Y     => Y,
              INCY  => 1,
              A     => A,
              LDA   => A'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR2;

   ----------
   -- SYRK --
   ----------

   procedure SSYRK (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, SSYRK, Name_Prepend & "ssyrk" & Name_Append);

   procedure DSYRK (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, DSYRK, Name_Prepend & "dsyrk" & Name_Append);

   procedure SYRK (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < N or C'Length (2) < N then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if A'Length (1) < N or A'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            SSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            DSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYRK;

   procedure SYRK (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type
   ) is
      K : Fortran_Integer;
   begin
      if A'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) /= C'Length (2) then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;
      
      case Precision is
         when Single =>
            SSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            DSYRK (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYRK;

   -----------
   -- SYR2K --
   -----------

   procedure SSYR2K (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, SSYR2K, Name_Prepend & "ssyr2k" & Name_Append);

   procedure DSYR2K (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     B     : in     Matrix_Type;
     LDB   : in     Fortran_Integer;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type;
     LDC   : in     Fortran_Integer
   );
   pragma Import (Fortran, DSYR2K, Name_Prepend & "dsyr2k" & Name_Append);

   procedure SYR2K (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     N     : in     Natural;
     K     : in     Natural;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     B     : in     Matrix_Type;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        C'Length (1) < N or C'Length (2) < N then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if A'Length (1) < N or A'Length (2) < K or
              B'Length (1) < N or B'Length (2) < K then
               raise Argument_Error;
            end if;
         when Transpose | Conjugate_Transpose =>
            if A'Length (1) < K or A'Length (2) < N or
              B'Length (1) < K or B'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            SSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            DSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR2K;

   procedure SYR2K (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     ALPHA : in     Float_Type'Base;
     A     : in     Matrix_Type;
     B     : in     Matrix_Type;
     BETA  : in     Float_Type'Base;
     C     : in out Matrix_Type
   ) is
      K : Fortran_Integer;
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or C'Length (1) = 0 or
        A'Length (1) /= B'Length (1) or A'Length (2) /= B'Length (2) or
          C'Length (1) /= C'Length (2) then
         raise Argument_Error;
      end if;
      
      case TRANS is
         when None =>
            if A'Length (1) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (2);
         when Transpose | Conjugate_Transpose =>
            if A'Length (2) /= C'Length (1) then
               raise Argument_Error;
            end if;
            K := A'Length (1);
      end case;
      
      case Precision is
         when Single =>
            SSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Double =>
            DSYR2K (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              N     => C'Length (1),
              K     => K,
              ALPHA => ALPHA,
              A     => A,
              LDA   => A'Length (1),
              B     => B,
              LDB   => B'Length (1),
              BETA  => BETA,
              C     => C,
              LDC   => C'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end SYR2K;
   
   ----------
   -- TBMV --
   ----------

   procedure STBMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, STBMV, Name_Prepend & "stbmv" & Name_Append);

   procedure DTBMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DTBMV, Name_Prepend & "dtbmv" & Name_Append);

   procedure TBMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) <= K or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            DTBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBMV;

   procedure TBMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) /= K + 1 or X'Length /= A'Length (2) or
        K >= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            DTBMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBMV;

   ----------
   -- TBSV --
   ----------

   procedure STBSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, STBSV, Name_Prepend & "stbsv" & Name_Append);

   procedure DTBSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     K     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DTBSV, Name_Prepend & "dtbsv" & Name_Append);

   procedure TBSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) <= K or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            DTBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBSV;

   procedure TBSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     K     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) /= K + 1 or X'Length /= A'Length (2) or
        K >= A'Length (2) then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            DTBSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (2),
              K     => Fortran_Integer (K),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TBSV;

   ----------
   -- TPMV --
   ----------

   procedure STPMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, STPMV, Name_Prepend & "stpmv" & Name_Append);

   procedure DTPMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DTPMV, Name_Prepend & "dtpmv" & Name_Append);

   procedure TPMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or
        (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            DTPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPMV;

   procedure TPMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (1 + X'Length) / 2 then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Double =>
            DTPMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPMV;

   ----------
   -- TPSV --
   ----------

   procedure STPSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, STPSV, Name_Prepend & "stpsv" & Name_Append);

   procedure DTPSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DTPSV, Name_Prepend & "dtpsv" & Name_Append);

   procedure TPSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     AP    : in     Vector_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if AP'Length < N * (N + 1) / 2 or INCX = 0 or
        (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            DTPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              AP    => AP,
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPSV;

   procedure TPSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     AP    : in     Vector_Type;
     X     : in     Vector_Type
   ) is
   begin
      if AP'Length /= X'Length * (1 + X'Length) / 2 then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Double =>
            DTPSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => X'Length,
              AP    => AP,
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TPSV;

   ----------
   -- TRMM --
   ----------

   procedure STRMM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, STRMM, Name_Prepend & "strmm" & Name_Append);

   procedure DTRMM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, DTRMM, Name_Prepend & "dtrmm" & Name_Append);

   procedure TRMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;
      
      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            STRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            DTRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMM;

   procedure TRMM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        A'Length (1) /= A'Length (2) then
         raise Argument_Error;
      end if;
      
      case SIDE is
         when Left =>
            if A'Length (1) /= B'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= B'Length (2) then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            STRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            DTRMM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMM;

   ----------
   -- TRMV --
   ----------

   procedure STRMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, STRMV, Name_Prepend & "strmv" & Name_Append);

   procedure DTRMV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DTRMV, Name_Prepend & "dtrmv" & Name_Append);

   procedure TRMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            DTRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMV;

   procedure TRMV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        A'Length (1) /= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            DTRMV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRMV;

   ----------
   -- TRSM --
   ----------

   procedure STRSM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, STRSM, Name_Prepend & "strsm" & Name_Append);

   procedure DTRSM (
     SIDE   : in     Character_Set;
     UPLO   : in     Character_Set;
     TRANSA : in     Character_Set;
     DIAG   : in     Character_Set;
     M      : in     Fortran_Integer;
     N      : in     Fortran_Integer;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     LDA    : in     Fortran_Integer;
     B      : in     Matrix_Type;
     LDB    : in     Fortran_Integer
   );
   pragma Import (Fortran, DTRSM, Name_Prepend & "dtrsm" & Name_Append);

   procedure TRSM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     M      : in     Natural;
     N      : in     Natural;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        B'Length (1) < M or B'Length (2) < N then
         raise Argument_Error;
      end if;
      
      case SIDE is
         when Left =>
            if A'Length (1) < M or A'Length (2) < M then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) < N or A'Length (2) < N then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            STRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            DTRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => Fortran_Integer (M),
              N      => Fortran_Integer (N),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSM;

   procedure TRSM (
     SIDE   : in     Side_Type;
     UPLO   : in     Triangle_Type;
     TRANSA : in     Transpose_Type;
     DIAG   : in     Diagonal_Type;
     ALPHA  : in     Float_Type'Base;
     A      : in     Matrix_Type;
     B      : in     Matrix_Type
   ) is
   begin
      if A'Length (1) = 0 or B'Length (1) = 0 or
        A'Length (1) /= A'Length (2) then
         raise Argument_Error;
      end if;
      
      case SIDE is
         when Left =>
            if A'Length (1) /= B'Length (1) then
               raise Argument_Error;
            end if;
         when Right =>
            if A'Length (1) /= B'Length (2) then
               raise Argument_Error;
            end if;
      end case;
      
      case Precision is
         when Single =>
            STRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Double =>
            DTRSM (
              SIDE   => Fortran_SIDE (SIDE),
              UPLO   => Fortran_UPLO (UPLO),
              TRANSA => Fortran_TRANS (TRANSA),
              DIAG   => Fortran_DIAG (DIAG),
              M      => B'Length (1),
              N      => B'Length (2),
              ALPHA  => ALPHA,
              A      => A,
              LDA    => A'Length (1),
              B      => B,
              LDB    => B'Length (1)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSM;

   ----------
   -- TRSV --
   ----------

   procedure STRSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, STRSV, Name_Prepend & "strsv" & Name_Append);

   procedure DTRSV (
     UPLO  : in     Character_Set;
     TRANS : in     Character_Set;
     DIAG  : in     Character_Set;
     N     : in     Fortran_Integer;
     A     : in     Matrix_Type;
     LDA   : in     Fortran_Integer;
     X     : in     Vector_Type;
     INCX  : in     Fortran_Integer
   );
   pragma Import (Fortran, DTRSV, Name_Prepend & "dtrsv" & Name_Append);

   procedure TRSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     N     : in     Natural;
     A     : in     Matrix_Type;
     X     : in     Vector_Type;
     INCX  : in     Integer
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) < N or A'Length (2) < N or
        INCX = 0 or (N - 1) * abs INCX >= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Double =>
            DTRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => Fortran_Integer (N),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => Fortran_Integer (INCX)
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSV;

   procedure TRSV (
     UPLO  : in     Triangle_Type;
     TRANS : in     Transpose_Type;
     DIAG  : in     Diagonal_Type;
     A     : in     Matrix_Type;
     X     : in     Vector_Type
   ) is
   begin
      if A'Length (1) = 0 or A'Length (1) /= A'Length (2) or
        A'Length (1) /= X'Length then
         raise Argument_Error;
      end if;

      case Precision is
         when Single =>
            STRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Double =>
            DTRSV (
              UPLO  => Fortran_UPLO (UPLO),
              TRANS => Fortran_TRANS (TRANS),
              DIAG  => Fortran_DIAG (DIAG),
              N     => A'Length (1),
              A     => A,
              LDA   => A'Length (1),
              X     => X,
              INCX  => 1
            );
         when Unsupported =>
            raise Unsupported_Precision_Error;
      end case;
   end TRSV;

end Ada_BLAS.Real;

